{-# OPTIONS --safe #-}

module Demo where

open import Data.List.Base
open import Data.Nat.Base hiding (_+_)
open import Data.Fin.Base hiding (_+_)

open import Relation.Binary.PropositionalEquality
open import Relation.Nullary

open import Generics

data Nat : Set where
  zero : Nat
  succ : Nat → Nat

natD : HasDesc Nat
natD = deriveDesc Nat

elimNat : (P  : Nat → Set) (P0 : P zero) (PS : ∀ {n} → P n → P (succ n))
        → ∀ n → P n
elimNat = deriveElim natD

elimNat-succ : ∀ P P0 (PS : ∀ {n} → P n → P (succ n)) n
             → elimNat P P0 PS (succ n) ≡ PS (elimNat P P0 PS n)
elimNat-succ P P0 PS n = refl

_+_ : Nat → Nat → Nat
_+_ n = elimNat _ n succ


decNat : DecEq Nat
decNat = deriveDecEq natD

-- vs

instance
  decNat' : DecEq Nat
  decNat' ._≟_ zero     zero     = yes refl
  decNat' ._≟_ zero     (succ y) = no λ ()
  decNat' ._≟_ (succ x) zero     = no λ ()
  decNat' ._≟_ (succ x) (succ y) with x ≟ y
  ... | yes refl = yes refl
  ... | no p     = no (λ { refl → p refl })


data Tree  (A : Set) : Set where
  node : A → List (Tree A) → Tree A

data Tree' (A : Set) : Set where
  node : A → (∀ {n} → Fin n → Tree' A) → Tree' A

data Thing (A : Set) : Set where
  cons : Thing A → Thing A

ThingD : HasDesc Thing
ThingD = deriveDesc Thing
